/*
 * express18.c
 * Copyright (C) 2006-2013, 2017, Ciprian Niculescu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <config.h>

#if defined HAVE_FEATURES_H
# include <features.h>
#endif				/* HAVE_FEATURES_H */

#include <getopt.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <aime.h>
#include <cardinal-wx.h>
#include <float1.h>
#include <inter.h>
#include <lcardinal.f.h>
#include <lintegral.f.h>
#if __EXPONENTIAL_REAL_SYNTAX__
# include <real.e.h>
#else
# include <real.h>
#endif				/* __EXPONENTIAL_REAL_SYNTAX__ */
#include <vkeylook-l.h>
#include <vtkey.h>

#define __HAIRY_INTERFACE__		0
#define __STRAIGHT_INTERFACE__		1

#define __INTERFACE_TEST__		__HAIRY_INTERFACE__

#define I_BILL(i)			(*((X1f4_E4_C_BILL *) (i)))
#define I_MODE(i)			(*((X1f4_E4_C_MODE *) (i)))
#define I_REAL(r)			(*((X1f4_E4_C_REAL *) (r)))
#define I_TEXT(t)			(*((X1f4_E4_C_TEXT *) (t)))

typedef struct context_type {
    struct x1f4_variable_type *variable_data;
#if __INTERFACE_TEST__ == __HAIRY_INTERFACE__
    struct x1f4_vkeytree_type *variable_tree;
#endif				/* __INTERFACE_TEST__ == __HAIRY_INTERFACE__ */
    void **state;
} context_type;

typedef union state_type {
    X1f4_E4_C_MODE mode;
    X1f4_E4_C_REAL real;
    X1f4_E4_C_TEXT text;
} state_type;

#if __INTERFACE_TEST__ == __HAIRY_INTERFACE__
static int init_attributes(struct x1f4_attributes_type *);
static int init_functions(struct x1f4_attributes_type *);
static int init_variables(struct x1f4_attributes_type *);
#endif				/* __INTERFACE_TEST__ == __HAIRY_INTERFACE__ */
static int select_function(const char *, unsigned, const void *,
			   const struct x1f4_function_type **);
static int select_variable(const char *, unsigned, const void *,
			   const struct x1f4_variable_type **, void **);
static int test_cast(void *, const char *, unsigned);

#if __INTERFACE_TEST__ == __HAIRY_INTERFACE__
static void fini_attributes(struct x1f4_attributes_type *);
static void fini_functions(struct x1f4_attributes_type *);
static void fini_variables(struct x1f4_attributes_type *);
#endif				/* __INTERFACE_TEST__ == __HAIRY_INTERFACE__ */
static void usage(void);

#if __INTERFACE_TEST__ == __HAIRY_INTERFACE__
static int
init_attributes(struct x1f4_attributes_type *attributes_data)
{
    int status = 1;

    do {
	if (!init_functions(attributes_data)) {
	    if (!init_variables(attributes_data)) {
		status = 0;
		break;
	    }

	    fini_functions(attributes_data);
	}
    } while (0);

    return status;
}
#endif				/* __INTERFACE_TEST__ == __HAIRY_INTERFACE__ */


#if __INTERFACE_TEST__ == __HAIRY_INTERFACE__
static int
init_functions(struct x1f4_attributes_type *attributes_data)
{
    int status;
    const struct x1f4_function_type *function_data;
    struct x1f4_vkeytree_type *root;

    function_data = x1f4_e4_defaults;
    status = x1f4_make_vtkeyroot(&root, function_data->name, function_data);
    if (status) {
	status = X1f4_E4_ALLOC_ERROR;
    } else {
	function_data++;
	while (function_data->name) {
	    status = x1f4_make_vtkeynode
		(&root, function_data->name, function_data);
	    if (status) {
		status = X1f4_E4_ALLOC_ERROR;
		break;
	    } else {
		function_data++;
	    }
	}
    }

    if (status) {
	x1f4_free_vtkeydata(root);
    } else {
	attributes_data->function_set.context = root;
    }

    return status;
}
#endif				/* __INTERFACE_TEST__ == __HAIRY_INTERFACE__ */


#if __INTERFACE_TEST__ == __HAIRY_INTERFACE__
static int
init_variables(struct x1f4_attributes_type *attributes_data)
{
    int status;
    struct context_type *context_data;
    const struct x1f4_variable_type *variable_data;

    context_data = (void *) attributes_data->variable_set.context;
    variable_data = context_data->variable_data;
    if (!context_data->variable_data) {
	context_data->variable_tree = NULL;
	status = 0;
    } else {
	struct x1f4_vkeytree_type *root;

	status = x1f4_make_vtkeyroot
	    (&root, variable_data->name, variable_data);
	if (status) {
	    status = X1f4_E4_ALLOC_ERROR;
	} else {
	    variable_data++;
	    while (variable_data->name) {
		status = x1f4_make_vtkeynode
		    (&root, variable_data->name, variable_data);
		if (status) {
		    status = X1f4_E4_ALLOC_ERROR;
		    break;
		} else {
		    variable_data++;
		}
	    }
	}

	if (status) {
	    x1f4_free_vtkeydata(root);
	} else {
	    ((struct context_type *)
	     attributes_data->variable_set.context)->variable_tree = root;
	}
    }

    return status;
}
#endif				/* __INTERFACE_TEST__ == __HAIRY_INTERFACE__ */


static int
select_function(const char *f, unsigned length, const void *context,
		const struct x1f4_function_type **function)
{
#if __INTERFACE_TEST__ == __HAIRY_INTERFACE__
    int status = X1f4_E4_PARSE_ERROR;
    const void *key;

    if (!x1f4_look_vtkey_l(f, length, context, &key)) {
	status = 0;
	*function = key;
    }

    return status;
#else
    int status = X1f4_E4_PARSE_ERROR;
    const struct x1f4_function_type *function_data;

    function_data = context;
    if (function_data) {
	while (function_data->name) {
	    if (length == function_data->length
		&& !memcmp((void *) f, function_data->name, length)) {
		break;
	    }
	    function_data++;
	}
	if (function_data->name) {
	    status = 0;
	    *function = function_data;
	}
    }

    return status;
#endif				/* __INTERFACE_TEST__ == __HAIRY_INTERFACE__ */
}


static int
select_variable(const char *f, unsigned length, const void *context,
		const struct x1f4_variable_type **variable, void **state)
{
#if __INTERFACE_TEST__ == __HAIRY_INTERFACE__
    int status = X1f4_E4_PARSE_ERROR;
    const struct context_type *context_data;

    context_data = context;
    if (context_data->variable_tree) {
	const void *key;

	if (!x1f4_look_vtkey_l(f, length, context_data->variable_tree, &key)) {
	    status = 0;
	    *state = context_data->state
		[(struct x1f4_variable_type *) key
		 - context_data->variable_data];
	    *variable = key;
	}
    }

    return status;
#else
    int status = X1f4_E4_PARSE_ERROR;
    const struct context_type *context_data;
    const struct x1f4_variable_type *variable_data;

    context_data = context;
    variable_data = context_data->variable_data;
    if (variable_data) {
	while (variable_data->name) {
	    if (length == variable_data->length
		&& !memcmp((void *) f, variable_data->name, length)) {
		break;
	    }
	    variable_data++;
	}
	if (variable_data->name) {
	    status = 0;
	    *state = context_data->state
		[variable_data - context_data->variable_data];
	    *variable = variable_data;
	}
    }

    return status;
#endif				/* __INTERFACE_TEST__ == __HAIRY_INTERFACE__ */
}


static int
test_cast(void *cast, const char *name, unsigned size)
{
    return 1 ^ fwrite(name, size, 1, cast);
}


#if __INTERFACE_TEST__ == __HAIRY_INTERFACE__
static void
fini_attributes(struct x1f4_attributes_type *attributes_data)
{
    fini_variables(attributes_data);
    fini_functions(attributes_data);
}
#endif				/* __INTERFACE_TEST__ == __HAIRY_INTERFACE__ */


#if __INTERFACE_TEST__ == __HAIRY_INTERFACE__
static void
fini_functions(struct x1f4_attributes_type *attributes_data)
{
    x1f4_free_vtkeydata((void *) attributes_data->function_set.context);
}
#endif				/* __INTERFACE_TEST__ == __HAIRY_INTERFACE__ */


#if __INTERFACE_TEST__ == __HAIRY_INTERFACE__
static void
fini_variables(struct x1f4_attributes_type *attributes_data)
{
    struct x1f4_vkeytree_type *variable_tree;

    variable_tree = ((struct context_type *)
	attributes_data->variable_set.context)->variable_tree;
    if (variable_tree) {
	x1f4_free_vtkeydata(variable_tree);
    }
}
#endif				/* __INTERFACE_TEST__ == __HAIRY_INTERFACE__ */


static void
usage(void)
{
    puts("Usage: express18 [OPTIONS] EXPRESSION [NAME VALUE]\n\
Evaluate EXPRESSION.\n\
\n\
Options:\n\
  -P, --print			print expression\n\
  -a, --output-base BASE	set output base for integral results to BASE\n\
  -p, --precision DIGITS	set precision for printing reals to DIGITS\n\
  -t, --type			print expression type\n\
      --list			list available functions and exit\n\
      --help			display this help and exit\n\
      --version			output version information and exit");
}


int
main(int argc, char **argv)
{
    int do_print = 0, list_functions = 0, status = 0;
    unsigned output_base = 10, precision = 16;

    unsetenv("POSIXLY_CORRECT");

    {
	char *output_base_specs = NULL, *precision_specs = NULL;

	while (1) {
	    char c;
	    static struct option long_options[] = {
/* *INDENT-OFF* */
		{   "help",	    0x00,   NULL,   0x68    },
		{   "list",	    0x00,   NULL,   0x01    },
		{   "output-base",  0x01,   NULL,   0x61    },
		{   "version",	    0x00,   NULL,   0x76    },
		{   "precision",    0x01,   NULL,   0x70    },
		{   "print",	    0x00,   NULL,   0x50    },
		{   NULL,	    0x00,   NULL,   0x00    }
/* *INDENT-ON* */
	    };

	    c = getopt_long(argc, argv, "Pa:p:t", long_options, NULL);

	    if (!~c) {
		break;
	    }

	    switch (c) {
	    case 001:
		list_functions = 1;
		break;
	    case 'P':
		do_print = 1;
		break;
	    case 'a':
		output_base_specs = optarg;
		break;
	    case 'p':
		precision_specs = optarg;
		break;
	    case 't':
		do_print = 2;
		break;
	    case 'h':
		usage();

		return 0;
	    case 'v':
		printf("%s (%s) %s\n", argv[0], PACKAGE, VERSION);

		return 0;
	    case '?':
		return 1;
	    }
	}

	if (list_functions | do_print) {
	    output_base_specs = NULL;
	    precision_specs = NULL;
	}

	if (output_base_specs) {
	    if (x1f4_parse_wxcardinal
		(&output_base, output_base_specs, NULL, 0)) {
		fprintf(stderr, "%s: cannot parse output base specification:"
			" `%s'\n", argv[0], output_base_specs);

		return 1;
	    } else {
		if (output_base < 2 || 36 < output_base) {
		    fprintf(stderr, "%s: output base (%u) is out of range (0"
			    " .. 36)\n", argv[0], output_base);

		    return 1;
		}
	    }
	}

	if (precision_specs) {
	    if (x1f4_parse_wxcardinal(&precision, precision_specs, NULL, 0)) {
		fprintf(stderr, "%s: cannot parse precision specification: `"
			"%s'\n", argv[0], precision_specs);

		return 1;
	    }
	}
    }

    if (list_functions) {
	x1f4_print_functions(stdout, x1f4_e4_defaults);
    } else if (!((argc - optind) & 1)) {
	status = 1;
	fprintf(stderr, "%s: wrong number of arguments\nType `%s --help' for"
		" more information.\n", argv[0], argv[0]);
    } else {
	do {
	    const char *expression;
	    struct x1f4_eerecord_type eerecord;
	    struct x1f4_variable_type *variable_data;
	    unsigned count;
	    void **state = NULL, *x1f4_expression;

	    status = 1;

	    expression = argv[optind];
	    optind++;

	    count = (argc - optind) >> 1;
	    if (!count) {
		variable_data = NULL;
	    } else {
		variable_data = (struct x1f4_variable_type *)
		    malloc((sizeof(struct x1f4_variable_type) + sizeof(void *)
			    + sizeof(union state_type)) * count
			   + sizeof(struct x1f4_variable_type));
		if (!variable_data) {
		    break;
		} else {
		    struct x1f4_variable_type *slide1;
		    union state_type *slide3;
		    unsigned i;
		    void **slide2;

		    slide1 = variable_data;
		    slide2 = (void **) (slide1 + count + 1);
		    slide3 = (union state_type *) (slide2 + count);

		    state = slide2;

		    for (i = count; i; i--) {
			status = libx1f4i0_miss_list
			    (argv[0], argv[optind], count - i, variable_data);
			if (status) {
			    break;
			}

			slide1->name = argv[optind];
			slide1->length = strlen(argv[optind]);

			optind++;

			{
			    double real;

#if __EXPONENTIAL_REAL_SYNTAX__
			    status = x1f4_parse_ereal
				(&real, argv[optind], NULL);
#else
			    status = x1f4_parse_real
				(&real, argv[optind], NULL);
#endif				/* __EXPONENTIAL_REAL_SYNTAX__ */
			    if (status) {
				fprintf(stderr, "%s: cannot parse value spec"
					"ification for variable %u: `%s'\n",
					argv[0], count - i, argv[optind]);
				break;
			    }

			    slide3->real = real;
			}

			optind++;

			slide1->type = X1f4_E4_REAL;
			slide1->flags = 0;

			*slide2 = slide3;

			slide1++;
			slide2++;
			slide3++;
		    }
		    if (i) {
			free(variable_data);
			status = 1;
			break;
		    } else {
			slide1->name = NULL;
		    }
		}
	    }

	    {
		struct context_type context;
		struct x1f4_attributes_type attributes;

		context.state = state;
		context.variable_data = variable_data;

		attributes.bcollect_set.eerecord_data = &eerecord;
		attributes.function_set.get = select_function;
#if __INTERFACE_TEST__ == __STRAIGHT_INTERFACE__
		attributes.function_set.context = x1f4_e4_defaults;
#endif				/* __INTERFACE_TEST__ == __STRAIGHT_INTER... */
		x1f4_llink_operator1s(&attributes.operator1s);
		x1f4_llink_operator2s(&attributes.operator2s);
		attributes.terminator = 0;
		attributes.variable_set.get = select_variable;
		attributes.variable_set.context = &context;

#if __INTERFACE_TEST__ == __HAIRY_INTERFACE__
		do {
		    status = init_attributes(&attributes);
		    if (status) {
			break;
		    }
#endif				/* __INTERFACE_TEST__ == __HAIRY_INTERFACE__ */

		    status = x1f4_init_expression
			(&x1f4_expression, expression,
			 X1f4_E4_BCOLLECT | X1f4_E4_DDACCESS, &attributes);

#if __INTERFACE_TEST__ == __HAIRY_INTERFACE__
		    fini_attributes(&attributes);

		} while (0);
#endif				/* __INTERFACE_TEST__ == __HAIRY_INTERFACE__ */
	    }
	    if (status) {
		if (status == X1f4_E4_ALLOC_ERROR) {
		    perror(argv[0]);
		} else {
		    fprintf(stderr, "%s: cannot parse `%s'\n", argv[0],
			    expression);
		    fprintf(stderr, "%s: ", argv[0]);
		    x1f4_stat_expression(stderr, test_cast, &eerecord, NULL);
		    fprintf(stderr, "\n");
		}
	    } else {
		if (do_print) {
		    if (do_print == 2) {
			unsigned type;

			type = x1f4_type_expression(x1f4_expression);
			if (type == X1f4_E4_MODE) {
			    printf("mode\n");
			} else if (type == X1f4_E4_REAL) {
			    printf("real\n");
			} else if (type == X1f4_E4_TEXT) {
			    printf("text\n");
			} else {
			    printf("void\n");
			}
		    } else {
			x1f4_print_expression(stdout, x1f4_expression, 0);
			putchar('\n');
		    }
		} else {
		    union state_type output;

		    status = x1f4_link_expression(x1f4_expression, &output);
		    if (status) {
			fprintf(stderr, "%s: cannot evaluate `%s'\n", argv[0],
				expression);
		    } else {
			unsigned type;

			type = x1f4_type_expression(x1f4_expression);
			if (type == X1f4_E4_BILL) {
			    x1f4_fprint_lcardinal
				(stdout, 0, output_base, I_BILL(&output));
			    putchar('\n');
			} else if (type == X1f4_E4_MODE) {
			    x1f4_fprint_lintegral
				(stdout, 0, output_base, I_MODE(&output));
			    putchar('\n');
			} else if (type == X1f4_E4_REAL) {
			    x1f4_fprint_wfloat
				(stdout, 0, 0, precision, I_REAL(&output));
			    putchar('\n');
			} else if (type == X1f4_E4_TEXT) {
			    fprintf(stderr, "%s: `%s' evaluates to text\n",
				    argv[0], expression);
			} else {
			    fprintf(stderr, "%s: `%s' evaluates to void\n",
				    argv[0], expression);
			}
		    }
		}

		x1f4_fini_expression(&x1f4_expression);
	    }

	    if (variable_data) {
		free(variable_data);
	    }
	} while (0);
    }

    return status;
}
